<?php namespace awp;

class ConfigPageOptions
{
	public $slug;	//string (url friendly) ?page=slug, page name in the url
	public $parent; //Object(ConfigPage) or string
	public $title;	//string, displays at top of page
	public $menuTitle; 		//if not null: sets the menu title. else: page wont be navagateable
	public $subMenuTitle; 	//if not null: sets the sub menu title, can be different to menu title, wont show in sub menu
	public $tabTitle;		//if not null: sets the tab title in page. else: tab wont be added to tab list
}

abstract class ConfigPage extends Module
{

	public function __construct($pageInfo, $styleHandles = array(), $scriptHandles=array())
	{
		parent::__construct();
		

		if(!$this->RequreKeys($pageInfo, "slug", "title", "menuTitle"))
		{
			echo "Config Module Requires (object/array)pageInfo to include, these properties: slug, title and menuTitle<br/>";
			var_dump($pageInfo);
			die();
			return;
		}

		$this->formData = array();
		$this->formOutput = array();
		
		$this->pageTabs = array();

		$this->pageInfo = cast ($pageInfo, "\awp\ConfigPageOptions");
		$this->parent = null;
		
		$this->requestData = $this->RequestData();
		
		
		//setup parent for page tabs and skip init.
		if(gettype($this->pageInfo->parentSlug) == "object" && $this->pageInfo->parentSlug instanceof \awp\ConfigPage)
		{
			$this->parent =  $this->pageInfo->parentSlug;
			$this->pageInfo->parentSlug = $this->parent->pageInfo->slug;
			
        	$this->parent->addTab($this);
			
			return;	
		}
		
		add_action("plugins_loaded", function()use($styleHandles, $scriptHandles) {
			$this->Init($styleHandles, $scriptHandles);
		});
		
		
	}

	/**
	 * Called to initialise data before page loads, usually setup database and fetch field data. Is called before both load and request.
	 **/ protected abstract function Begin();

	/**
	 * Called before render page
	 **/ protected abstract function Load();

	/**
	 * Called when data is posted.
	 **/ protected abstract function Request($data);



	public function Init($styleHandles = array(), $scriptHandles = array())
	{
		
		
		//if tab is selected call that init instead.
		if(!$this->CurrentTab())
		{
			if($findTab = $this->FindTab())
			{
				$findTab->Init($styleHandles, $scriptHandles);
				return;	
			} 
		}
		//var_dump($this->pageInfo->slug);
		//var_dump($this->CurrentPage());
		//echo "<br/>";
		
		add_action( "admin_menu", function (){ $this->CreatePage(); } );

		if( !$this->CurrentPage())
			return;
			
		add_action("init", function() { $this->Begin(); });

		$this->RequireMod("form");
		$this->RequireMod("admin-pages");
		$this->RequireMod("debug");


		$this->styles = $styleHandles ;
		$this->scripts = $scriptHandles;
		
		$this->Load();
		add_action( 'admin_init', function() { $this->BuildForms();} );

		add_action("admin_enqueue_scripts", function(){
			wp_register_style( 'style-forms', AWP_URL . "styles/forms.css", array(), null );
			wp_register_style( 'style-buttons',  AWP_URL . "styles/buttons.css", array(), null );
			wp_register_style( 'style-cards',  AWP_URL . "styles/cards.css", array(), null );
			wp_register_style( 'style-tables',  AWP_URL . "styles/tables.css", array(), null );

		});
	}

	/**
	 * Add a component to page, either form or html element.
	 **/ 
	protected function AddComponent($formData)
	{
		
			
		if(!$this->CurrentPage())
			return;
		

		$formDataExample = array(
			"form1" => array(
				"group1" => array(
					"groupTitle" => "group title",
					"field1" => array( "label"=>"",  "default"=>"",  "type"=>"", "description"=>"", ),
				),
				"submit" => array(
					"text" => "",
					"action" => "", // options or custom_action 
					"settingName" => "",
				)
			),
			
			//"content" => function(){ return myFunc() },
		);


		$this->ValidateFormData($formData);
		$this->formData =  array_merge ($this->formData, $formData);
		

		//CREATE SUBMIT PAGE...
		if(isset($_REQUEST["action"]))
		{
			//die();
			$formKeys = array_keys($formData);
			
			for($i=0; $i<count($formKeys); $i++)
			{
				$formKey = $formKeys[$i];
				$groups = $formData[$formKey];
				if(!is_array($groups))
					continue;

				$action = $groups["submit"]["action"];
				if($action != "options")
				{ 
					//loads on admin-post.php
					add_action( "admin_post_{$action}", function($input) use($groups){
						$this->HandleSubmition($groups["submit"]["action"]);
					}  );
				}

			}
		}
	}
	
	public function AddTab($page)
	{
		if($this != $page && !is_subclass_of(get_class($page), get_class()) )
		{
			return;
		}
		
		$this->pageTabs[] = $page;
	}


	/**
	 * Takes form data and makes sure it contains the correct properties, if not errors will show instructions.
	 **/
	private function ValidateFormData($formData)
	{
		$formKeys = array_keys($formData);
		for($i=0; $i<count($formKeys); $i++)
		{
			$formKey = $formKeys[$i];
			$groups = $formData[$formKey];

			if(!is_array($groups))
				continue;

			$groupKeys = array_keys($groups);

			if(!$this->RequreKeys($groups, "submit") && count($groups) <= 1)
			{
				echo "<h2>Config Module Form Data Invalid</h2><br/>Must contain a group called submit and 1 aditonal group with a unique_key.";
				die();
				return;
			}

			for($j=0; $j<count($groupKeys); $j++)
			{
				$groupKey = $groupKeys[$j];
				$fields = $groups[$groupKey];
				$fieldKeys = array_keys($fields);
				
				
				if($groupKey == "submit")
				{
					if(!$this->RequreKeys( $groups["submit"], "action", "settingName"))
					{
						echo "<h2>Config Module Form Data Invalid</h2><br/>Submit Group must contain the properties: action and settingName<br/>";
						var_dump($fields["submit"]);
						die();
						return;
					}
					continue;
				}

				for($k=0; $k<count($fieldKeys); $k++)
				{
					$fieldKey =  $fieldKeys[$k];
					$filed = $fields[$fieldKey];
					if(!$this->RequreKeys( $filed, "label", "default", "type"))
					{
						echo "<h2>Config Module Form Data Invalid</h2><br/>Field must contain the properties: label, default, and type.<br/>";
						var_dump($filed);
						die();
						return;
					}
				}


			}

		}
	}

	/**
	 * Makes sure the first object contains the following properties. 
	 **/
	private function RequreKeys($arrayOrObject, $string, $moreStrings=null)
	{
		//use insert keys and compare the 2 objects.
		return call_user_func_array(array($this,"InsertKeys"), func_get_args()) === $arrayOrObject;
	}

	/**
	 * Pads an object with empty strings, used for avoiding undefined checks.
	 **/
	private function InsertKeys($arrayOrObject, $string, $moreStrings=null)
	{
		$keys = func_get_args();
		$item = array_shift($keys);

		for($i=0; $i<count($keys); $i++)
		{
			$key = $keys[$i];
			if(is_array($item) && !isset($item[$key]))
			{
				$item[$key] = "";
			}
			else if(is_object($item) && !property_exists($item, $key))
			{
				$item->$key = "";
			}
		}
		
		return $item;
	}
	
	/**
	 * WP wrapper, creates the page in the admin area. 
	 **/
	private function CreatePage()
	{
		$pageHook = null;


		$capability = property_exists($this->pageInfo, "capability") ? $this->pageInfo->capability : "manage_options";
		$pagePosition = property_exists($this->pageInfo, "position") ? $this->pageInfo->position : null;
		$iconUrl = property_exists($this->pageInfo, "iconUrl") ? $this->pageInfo->iconUrl : "";

		$parentSlug = property_exists($this->pageInfo, "parentSlug") ? $this->pageInfo->parentSlug : null;
		$optionsParent = false;
		
		
		$settingSubPages = $GLOBALS["submenu"]["options-general.php"];
		$settingSubKeys = array_keys($settingSubPages);
		if( $parentSlug )
		{
			for($i=0; $i<count($settingSubKeys); $i++)
			{
				if($settingSubPages[$settingSubKeys[$i]][2] == $parentSlug)	
				{
					$optionsParent = true;
				}
			}
		}
		
		$newParent = $parentSlug !== false && !isset($GLOBALS["admin_page_hooks"][$this->pageInfo->parentSlug ]) && !$optionsParent;

		if($newParent)
		{
			$pageInfo = $this->parent ? $this->parent->pageInfo : $this->pageInfo;
			$pageHook = add_menu_page( $this->pageInfo->title, $this->pageInfo->menuTitle, $capability, $pageInfo->slug, function(){$this->BuildPage();}, $iconUrl, $pagePosition );
		}
		else if( !empty($parentSlug) && array_key_exists($parentSlug, $GLOBALS["admin_page_hooks"]) )
		{
			$pageHook = add_submenu_page($parentSlug, $this->pageInfo->title, $this->pageInfo->menuTitle, $capability, $this->pageInfo->slug, function(){this->BuildPage();}, $pagePosition );
		}
		else
		{
			//addds the page  to the options if cannot find or create parent 
			$pageHook = add_options_page( $this->pageInfo->title, $this->pageInfo->menuTitle, $capability, $this->pageInfo->slug,  function(){$this->BuildPage();}, $pagePosition);
		}

		//rename the sub-menu title. e.g. Posts is renamed to All Posts on the drop-down.
		if($newParent && property_exists($this->pageInfo, "subMenuTitle"))
		{
			add_submenu_page( $this->pageInfo->slug, $this->pageInfo->title, $this->pageInfo->subMenuTitle, $capability, $this->pageInfo->slug );
		}

		
		if(!$newParent && empty($this->pageInfo->menuTitle))
		{
			$currentSlug = $this->pageInfo->slug;
			add_action( 'current_screen', function() use($parentSlug, $currentSlug, $optionsParent){
				remove_submenu_page( $optionsParent ? "options-general.php" : $parentSlug, $currentSlug );
			}, 21 );
		}
		//add_action("load-{$pageHook}", function (){ $this->Load(); });

		add_action("admin_enqueue_scripts", function($hook)use($pageHook){
			if($hook !== $pageHook || !$this->CurrentPage())
				return;
			
			for($i=0; $i<count($this->styles); $i++)
			{
				wp_enqueue_style( $this->styles[$i] );
			}

			for($i=0; $i<count($this->scripts); $i++)
			{
				wp_enqueue_script( $this->scripts[$i] );
			}


		});
	}

	/**
	 * Generates the HTML for the page.
	 **/
	private function BuildPage()
	{
		
			$pageInfo = $this->parent ? $this->parent->pageInfo : $this->pageInfo;
			$pageTabs = $this->parent ? $this->parent->pageTabs : $this->pageTabs;
			//var_dump($pageTabs);
		?>
		<div class="wrap">
			<h1></h1><!--app messages are injected here after first H1--><hr/>
		<?php
		$currentTab =  isset($_REQUEST["tab"]) ? $_REQUEST["tab"] : null;
		if(count($pageTabs)): ?>
		<nav class="nav-tab-wrapper">
			<a href="?page=<?=$pageInfo->slug?>" class="nav-tab <?php if($currentTab==null || $currentTab == $pageInfo->slug ):?>nav-tab-active<?php endif; ?>"><?=$pageInfo->title?></a>
			
		    <?php for($i=0; $i<count($pageTabs); $i++){
		    	$itemTab = $pageTabs[$i]->pageInfo->slug;
		    	echo "<a href=\"?page={$pageInfo->slug}&tab={$itemTab}\" class=\"nav-tab " . ($currentTab == $itemTab ? "nav-tab-active" : "") . "\">{$pageTabs[$i]->pageInfo->title}</a>";
				
			}?>
		</nav>
	 	<div class="tab-content">
		<?php endif; ?>
			<?php for($i=0; $i<count($this->formOutput); $i++){
				echo $this->formOutput[$i];
			}?>
		<?php if(count($this->pageTabs)): ?></div><?php endif;?>
		</div>
		<?php
	}


	/**
	 * Loop form data and create and store each form as HTML. 
	 **/
	private function BuildForms()
	{
		$formKeys = array_keys($this->formData);

		for($i=0; $i<count($formKeys); $i++)
		{
			$formKey= $formKeys[$i];
			$groups = $this->formData[$formKey];

			if(is_callable($groups))
			{
				ob_start( ); 
					call_user_func($groups);
					$this->formOutput[] = ob_get_contents();
				ob_end_clean();
			}
			if(!is_array($groups))
				continue;

			$groupKeys = array_keys($groups);
			
			
			
			register_setting( $formKey, $groups["submit"]["settingName"], array("sanitize_callback"=>function($data)use($groups){ return $this->ParseInput($data, $groups);}) );
		
			for($j=0; $j<count($groupKeys); $j++)
			{
				$group = $groupKeys[$j];

				if($groupKeys[$j] != "submit")
				{
					$this->BuildForm($formKey, $group, $groups[$group], $groups["submit"]["settingName"]);

				}
			}

			$action = $groups["submit"]["action"];
			$submitText = isset($groups["submit"]["text"]) ? $groups["submit"]["text"] :  "";
			$submitText = empty($submitText) ? null : $submitText;


			//add_action("admin_init", function() use($formKey, $action){

				ob_start( ); 
					echo "<form method=\"POST\" id=\"{$formKey}\" action=\"". ($action == "options" ? "options.php" : "admin-post.php") . "\">"; 
					$this->app->IncludeFormHeader( $formKey, $action == "options" ? "update" : $action );
					//if(isset($_REQUEST["tab"]))
						//echo "<input type=\"hidden\" name=\"tab\" value=\"{$_REQUEST["tab"]}\" />";
					
					$this->app->DisplaySettingsSection( $formKey );
					submit_button($submitText);
					?></form><?php
					$this->formOutput[] = ob_get_contents();
				ob_end_clean();

			//});
		}
	}

	/**
	 * WP wrapper to create the page form.
	 **/
	private function BuildForm($form, $sectionName, $fieldData, $settingName)
	{
		//print_r($sectionName);echo "<br/>";
		$groupTitle = $sectionName;
		if(array_key_exists("groupTitle", $fieldData))
		{
			$groupTitle = $fieldData["groupTitle"];
			unset($fieldData["groupTitle"]);
		}

		//register_setting( $this->slug, $this->settingName, array("sanitize_callback"=>function($data)use($fieldData){ return $this->ParseInput($data, $fieldData);}) );
		add_settings_section( $sectionName, $groupTitle, null, $form );

		//loops through fieleds and loads the HTML into WP ready for page to load.
		$this->app->CreateFields($form , $fieldData, $settingName, $sectionName);
	}

	
	private function HandleSubmition($action)
	{
		//$this->Begin();
		$sanitisedInput = $this->ValidateFormSubmition();

		if(!$sanitisedInput)
			return;

		//Parse Types
		$formKeys = array_keys($this->formData);
		for($i=0; $i<count($formKeys); $i++)
		{
			$form = $this->formData[$formKeys[$i]];
			$groupKeys =  array_keys($form);

			for($j=0; $j<count($groupKeys); $j++)
			{
				$group = $form[$groupKeys[$j]];
				$inputKeys = array_keys($sanitisedInput);
				
				for($k=0; $k<count($inputKeys); $k++)
				{
					if(array_key_exists($inputKeys[$k], $group))
					{
						//print_r($group[$inputKeys[$k]]["type"]);echo "<br/>";
						$sanitisedInput[$inputKeys[$k]] = $this->app->CheckInput($group[$inputKeys[$k]]["type"], $sanitisedInput[$inputKeys[$k]]);
					}
				}

			}


		}


		$this->Request((object)$sanitisedInput);
	}

	protected function ValidateFormSubmition($permission = "")
	{
		$permission = $permission ==  "" ? "manage_options" : $permission;
		if( !current_user_can( $permission ) && wp_verify_nonce( $_POST["_wpnonce"], $_POST["option_page"] . "-options" ) )
		{
			$this->RedirectBack();
		}

		if(!array_key_exists($_REQUEST["option_page"], $this->formData))
			return null;


		$formData = $this->formData[$_REQUEST["option_page"]];
		$settingName = $formData["submit"]["settingName"];
		//todo: check security on inputs

		$inputKeys = array_keys($_REQUEST[$settingName]);
		$sanitisedInput = array();
		
		for($i=0; $i<count($inputKeys); $i++)
		{
			$inputKey = $inputKeys[$i];
			$inputValue = $_REQUEST[$settingName][$inputKey];

			//var_dump($inputValue);
			//echo "<br/>";

			if($inputKey == "slug")
				$sanitisedInput[$inputKey] = sanitize_key($inputValue);
			else
				$sanitisedInput[$inputKey] = sanitize_text_field($inputValue);
		}

		return $sanitisedInput;
	}

	/**
	 * Jump back to previous page
	 **/
	protected function RedirectBack($fallback = "")
	{
		$url  = isset($_REQUEST['_wp_http_referer']) ? $_REQUEST['_wp_http_referer'] : $fallback;
		
		header("Location: {$url}");
		die();
	}

	private function RequestData()
	{
		$queryData;
		$returnData = ["method"=>"request", "page"=>null, "tab"=>null];
		
		//fetch refer data if availible
		if(isset( $_REQUEST["_wp_http_referer"] ))
		{
			parse_str( parse_url( $_REQUEST["_wp_http_referer"], PHP_URL_QUERY), $queryData);
			$returnData["page"] = isset($queryData["page"]) ? $queryData["page"]: null;
			$returnData["tab"] = isset($queryData["tab"]) ? $queryData["tab"] : null;
			$returnData["method"] = "referer";
		}
		
		//overwrite the refer link with current page data
		$returnData["page"] = isset($_REQUEST["page"]) ? $_REQUEST["page"] : $returnData["page"];
		$returnData["tab"] = isset($_REQUEST["tab"]) ? $_REQUEST["tab"] : $returnData["tab"];
		
		//cant have tab without page
		$returnData["tab"] = $returnData["page"] !== $returnData["tab"] ? $returnData["tab"] : null;
		
		return (object) $returnData;
	}
	
	private function FindTab()
	{
		
		$requestPage = $this->requestData->page;
		$requestTab = $this->requestData->tab;
		if($requestTab)
		{	
			for($i=0; $i<count($this->pageTabs); $i++)
			{
				
				if ($this->pageTabs[$i]->pageInfo->slug == $requestTab && $requestPage  == $this->pageTabs[$i]->pageInfo->parentSlug)
				{
					return $this->pageTabs[$i];
				}
			}
		}
	}
	
	private function CurrentTab()
	{
		return $this->FindTab() === null;
		
	}
	
	private function CurrentPage()
	{

		$requestPage = $this->requestData->page;
		$requestTab = $this->requestData->tab;
		
		$getPage = isset($_REQUEST["page"]) ? $_REQUEST["page"] : null;
		$getTab = isset($_REQUEST["tab"]) ? $_REQUEST["tab"] : null;
		$getTab = $getPage !== $getTab ? $getTab : null;

		$slug = $this->pageInfo->slug;
		$findTab = $this->FindTab();
		
		
		
		$page = $requestPage === $slug ||( $requestTab  == $slug && $requestPage == $this->pageInfo->parentSlug);
		
		$request = isset($_REQUEST["action"]) && ($requestTab == $slug || $findTab && $findTab->pageInfo->parentSlug == $slug);
		
		$tab = empty($this->pageTabs) || $getTab  === null || $getTab === $slug || ( $findTab && $findTab->pageInfo->slug == $getTab) ;
		
		$options = basename($_SERVER['PHP_SELF']) === "options.php";
		
		
		/*
		var_dump($slug);echo "<br/>";
		if($page) echo"Page: {$_REQUEST['page']}<br/>";
		if($request) echo"request: {$_REQUEST["action"]}<br/>";
		if($tab) echo"tab: {$getTab}<br/>";
		if($options) echo"options<br/>";
		echo "<hr/>";
		
		//*/
		//if( ( $_REQUEST["page"] == $this->pageInfo->slug) || basename($_SERVER['PHP_SELF']) == "options.php")
		//$r =  ($request && $tab) || ($page && $tab)  || $options;
		//var_dump(($page && $tab));echo "<br/>";
		//var_dump($r);echo "{$this->pageInfo->slug}<br/>";
		
		
		//var_dump(($this->pageInfo));
		return ($request && $tab) || ($page && $tab)  || $options;

	}
	


	protected function ParseInput($inputData, $fieldData)
	{

        //NOTE this function can be called twice on the same input in options.php
        //once for add_option and once for update_option;
        //if you are paring types you may need to do extra checks to catch errors thrown on second pass.
        //alternatively you can add checks for the action and page and return $innuptData when you get add_option from options.php
		
		$output = array();
		
		if($inputData == null)
			$inputData = array();

		
		$groupKeys = array_keys($fieldData);

		for($i=0; $i<count($groupKeys); $i++)
		{

			if($groupKeys[$i] == "submit")
				continue;

			$group = $fieldData[$groupKeys[$i]];
			$fieldKeys = array_keys($group);

			for($j=0; $j<count($fieldKeys); $j++)
			{
				$fieldKey = $fieldKeys[$j];
				$fieldInfo = (object)$group[$fieldKey];

				if($fieldKey == "groupTitle")
				continue;
                
				$fieldInput = array_key_exists($fieldKey, $inputData) ? $inputData[$fieldKey]: null;

				if($fieldInput !== null)
				{	

					$fieldInput = $this->app->CheckInput($fieldInfo->type, $fieldInput);
				}
				else
				{
					$fieldInput = $fieldInfo->default;
				}

				$output[$fieldKey] = $fieldInput;

			}
		}

        //var_dump($output);
        //echo serialize (($output));
        //echo  json_encode($output);
		
		return $output;
	}







}


?>